MODULE StarField;
(* Graphics joke, XTDLIB independant except for CorePal, which is only
    necessary for 256 colors mode. A starfield entirely written from scratch,
    just for fun. Part of XtdLib package (C) by Marco van de Voort 1997

 A friend of my brother had this as a junior schoolassignment, and asked me
   how to do it. I couldn't tell directly, and made a test-project, here it is.
   The math appeared to be even simpler than I thought, no gonio,
   just one (!) formula with two congruent triangles. The more difficult part
   was estimating the size of the volume where the stars are created.
   (in procedure InitOneStar Universe[starnr].X,Y and Z).

 Personally I'm very satisfied with the result, but any other tricks could
  be added, specially to make it nicer when displaying little stars
  (on slower machines) many stars is much easier, since you always have a
  high coverage on your screen.

 I can think of several improvements :

  - All points in the creation universe are equally likely to give birth
     to stars. Distant points should be less likely, but some of it is
     already accounted for by making the creation area a volume instead of
     an plane. (less bright stars are visible later).

  - Real close stars could be made 4 or 5 pixels big:

                                            x
                                xx         xxx

                                xx          x

        I think the resolution 640x480x16 is high enough for it.
        (later : I implemented it for 640x480)

    - If you have a better graph module, you could change it to support
       more gray-scales, higher resolutions etc etc. Or use a nice X-mode
       module. Since the only functions used are SetVideoMode,TextMode and
       Plot, this shouldn't be really difficuly.

    - Or use the real distance (sqr(X^2+Y^2+Z^2) instead of Z.
       With the current coprocessor speeds, this could be possible.
       Or do calculations with a MMX processor on 250 stars at once,
       this is THE MMX application, many similar calculations with
       low precision :-)

    I disabled dX and dY, this gives a smoother starfield. A star in
    a quadrant moves always to the far border of the quadrant.

    When you use dX and dY, stars near the centre can change quadrant,
    which is kind of chaotic. Since stars don't move on this timescale,
    I disabled it but let the possibility exist, if somebody knows a better
    method. Or use a very low (1E-4) or so REAL value for dX and dY,
    and the scale the movement with Z, since start further away move more:-)

In theory, a star could hit you right in the face. If you observe this,
 please report! I wonder what it looks like.  *)

FROM Graph  IMPORT SetVideoMode,Plot,TextMode;
FROM SYSTEM IMPORT Eval;
FROM Lib    IMPORT RANDOM,RANDOMIZE,Delay;
FROM CorePal IMPORT RGBType,SetPalette;

IMPORT IO;

CONST MaxStar =190; (* # stars-1, easier than # stars in loops, because
                       I like to start at 0 with loops, an assembler left
                       over?
                       8191 stars (MaxStars=8190) is the maximum, because
                       of the 64K border. But my Cx686-166 is too slow....
                       4000 works nice, but is too full (I'm watching
                       Startrek Voyager right now).
                       Anything between 250 and 1000 will work, what range
                       is a matter of taste.

                       This constant is a matter of taste and machinespeed*)
      Distance=20.0; (* projection plane distance*)

      Col256  =FALSE; (* TRUE = 320x200x256 (smoother color changes)
                         FALSE= 640x480x16 (smoother movement and
                                            big pixels for near stars) *)

TYPE Star       = RECORD
                   X,Y,Z :INTEGER;  (* Coordinates, X and Y are like
                                        the screen resolution, Z is
                                        perpendicular to that, but the
                                        sign is always positive. *)

             (*      dX,dY, *)  dZ :INTEGER; (* Partial speed(s) *)
                                        (* If dZ is the only one used,
                                           you could regard dZ as the speed
                                           of a starship traveling thru
                                           the starfield *)
                   END;

     UniverseType   = ARRAY[0..MaxStar] OF Star;

VAR Universe    : UniverseType; (* The universe :-) *)
    DimX,DimY   : CARDINAL;     (* Dimensions of screen/2 *)

PROCEDURE SetInten;
(* Calculate a gray-scale palette. *)

VAR
  i      : CARDINAL;
  NP     : ARRAY[0..63] OF RGBType;

BEGIN
  FOR i := 0 TO 63 DO
    NP[i].Red   := BYTE(64-i);
    NP[i].Green := BYTE(64-i);
    NP[i].Blue  := BYTE(64-i);
    END;
   SetPalette(0,64, NP);
END SetInten;

PROCEDURE InitOneStar(St : CARDINAL);

BEGIN
  (* The size and place where the creation takes place.
      Quite critical to get a stable and diverse starfield.
     A nice change could be using a own random with relatively
     more stars in x,y = -2000..2000 than outside that region.
     Could also be done with a division factor?
     So the X and Y coordinates are divided by aninteger (range 1..3 or so)
     this puts relatively more stars in the centre without ruling the others
     out *)
  Universe[St].X:=INTEGER(RANDOM(20000))-10000;
  Universe[St].Y:=INTEGER(RANDOM(20000))-10000;
  Universe[St].Z:=INTEGER(RANDOM(500))+1;
  (* the color changing should be adapted if you change the .Z parameter *)

  (* Partial speeds, values are a matter of taste, but not too high *)
(*  Universe[St].dX:=INTEGER(RANDOM(7))-3;          (* -3 .. 3  *)
    Universe[St].dY:=INTEGER(RANDOM(7))-3;          (* -3 .. 3  *) (*#*) *)
  Universe[St].dZ:=RANDOM(3)+1;                   (*  1 .. 4  *)
END InitOneStar;

PROCEDURE BigBang;
(* Create the universe, create MaxStar+1 stars.  *)

VAR A :CARDINAL;

BEGIN
 FOR A:=0 TO MaxStar DO
  InitOneStar(A);
  END;
END BigBang;

PROCEDURE PlotStar(St:CARDINAL; Col:CARDINAL);

(* Project the star (coordinates in Universe) on a X-Y plane at distance 20
    This routine controls both the disappearing as the moving of a star
     if col=0 than the star disappears.
     if col=15 than the star is projected in a grayscale. *)


VAR RZ          : REAL;
    NX,NY       : LONGINT;
    C           : CARDINAL;
    BigPixels   : BOOLEAN;

BEGIN
 C:=Col;
 IF C=15 THEN  (* If a point is plotted (not erased with color 0),
                   than determine the grayscale as function of the
                   Z-distance/coordinate) *)
  IF Col256 THEN (* 256 color mode, Max distance= 550, divide in 64 steps *)
   C:=(CARDINAL(Universe[St].Z*64)) DIV 551;
   IF C>63 THEN
    C:=63;
    END;
  ELSE
   (* 16 color mode quite clear *)
   IF Universe[St].Z>350 THEN
    C:=8; (* far darkgray *)
   ELSIF Universe[St].Z>200 THEN
    C:=7; (* medium (light)gray *)
    END; (* else (close) white *)
   END;
  END;
BigPixels:=(NOT Col256) AND (Universe[St].Z<150);

(* Projection on a plane at distance 20, calculated with two congruent
    triangles (picture not on right scale, big triangle is much bigger
    and variable)

 Z     |\               <- observer, distance 0             z0 , x0 or y0
 |     | \
 V     |  \
       |---\            <- projection plane, distance 20    z20, x20 or y20
 A     |    \
 x     |     \
 i     |      \
 s     |-------\        <- distance Z                       Z  , X or Y

       X or Y axis ->

Assume X axis (y axis is analog)

    z-z0    z20-z0
    ----  = ------
    x-x0    x20-x0

z0=0 x0=0 z20=20    ->  x20=20/z * x
 *)

 RZ:=Distance/REAL(Universe[St].Z);
 NX:=LONGINT(RZ*REAL(Universe[St].X));
 NY:=LONGINT(RZ*REAL(Universe[St].Y));

 IF (ABS(NX)<LONGINT(DimX+1)) AND (ABS(NY)<LONGINT(DimY+1)) THEN
  (* projectable! -> plot it! *)
 Plot(DimX+CARDINAL(NX),DimY+CARDINAL(NY),C);
  IF BigPixels THEN
   Plot(DimX+1+CARDINAL(NX),DimY+CARDINAL(NY),C);
   Plot(DimX+CARDINAL(NX),DimY+1+CARDINAL(NY),C);
   Plot(DimX+1+CARDINAL(NX),DimY+1+CARDINAL(NY),C);
   END;

 ELSIF C#0 THEN
  InitOneStar(St); (* If projected star isn't visible and we're not
                       erasing stars then create a new one which will become
                       visible next turn. *)
  END;
END PlotStar;

PROCEDURE DoOneStep;

VAR A : CARDINAL;

BEGIN
 FOR A:=0 TO MaxStar DO        (* Do all stars. *)
  PlotStar(A,0);               (* erase last position *)

  (* calculate new one *)
  Universe[A].Z:=Universe[A].Z-INTEGER(Universe[A].dZ);
(*  Universe[A].Y:=Universe[A].Y+LONGINT(Universe[A].dY);
    Universe[A].X:=Universe[A].X+LONGINT(Universe[A].dX); *)

  IF (Universe[A].Z)<1 THEN
   InitOneStar(A);          (* If not visible anymore then create a new one*)
   END;
  PlotStar(A,15);           (* make it visible *)
  END;
END DoOneStep;


BEGIN
 RANDOMIZE;
 BigBang;                             (* Create initial stars *)
 IF Col256 THEN
  Eval(SetVideoMode(19));              (* 320x200x256 videomode *)
  DimX:=160;        DimY:=100;         (* 0,0 is middle of screen,
                                          so screen is -DimX..DimX *)
  SetInten;                            (* Calculate grayscale palette*)
 ELSE
  Eval(SetVideoMode(18));              (* 640x480x16 videomode *)
  DimX:=320;        DimY:=240;
  END;

 REPEAT
  DoOneStep;                          (* move the stars and create
                                         new ones when necessary *)
  Delay(50);                          (* machine independant, but should
                                         be lowered on slow machines.
                                         Setting it too low might cause
                                         flickering, you'd better decrease
                                         the amount of stars instead of
                                         this delay. *)
 UNTIL IO.KeyPressed();
 TextMode;                            (* Return to textmode, also restores
                                         palette when necessary. *)
END StarField.
